use std::process::exit;
use clap::Parser;

mod build_rule;
mod makefile;
mod rule_list;
use makefile::*;
use rule_list::*;

#[derive(Parser, Debug)]
struct Args {
    /// Replace default make output with nice build messages (colors!)
    #[arg(short, long)]
    pretty: bool,

    /// Generate a 'clean' rule that removes all build files
    #[arg(short, long)]
    clean: bool,

    /// Generate a 'run' rule that allows you run the executable with 'make run'
    #[arg(short, long)]
    run: bool,

    /// Configure to build a statically linked library
    #[arg(short, long)]
    static_library: bool,

    /// Configure to build a dynamically linked library
    #[arg(short, long)]
    dynamic_library: bool,

    /// Path to makefile
    #[arg(short, long, default_value = "Makefile")]
    makefile: String,
}

thread_local! {static ARGS: Args = Args::parse()}

fn main() {

    let mut makefile_name = String::new();
    ARGS.with(|args| {
        makefile_name.push_str(&args.makefile);
    });
    let makefile = Makefile::new_from_path(&makefile_name);

    let mut rule_list = RuleList::new();
    rule_list.recursively_gen_objects(".", &makefile).unwrap_or_else(|e| {
        panic!("error: problem in recursive search: {e}");
    });

    if rule_list.0.is_empty() {
        eprintln!("no source files found, nothing to do. exiting...");
        exit(0);
    }

    let mut generated_content = String::new();

    // Only include directory defaults if the user has not specified them
    let mut dirdefwritten = false;
    if ! makefile.contains_builddir_def() {
        generated_content.push_str("BUILDDIR = build\n");
        dirdefwritten = true;
    }

    if ! makefile.contains_objdir_def() {
        generated_content.push_str("OBJDIR = $(BUILDDIR)/obj\n");
        dirdefwritten = true;
    }

    if dirdefwritten {
        generated_content.push_str("\n");
    }



    rule_list.add_target_rule();

    rule_list.add_directory_rule("$(BUILDDIR)");
    rule_list.add_directory_rule("$(OBJDIR)");
    ARGS.with(|args| {
        if args.clean {
            rule_list.add_clean_rule();
        }
        if args.run {
            rule_list.add_run_rule();
        }
    });

    generated_content.push_str(&rule_list.to_string());

    // Write makefile
    makefile.write_new_generated_content(generated_content).unwrap_or_else(|e| {
        panic!("error: could not write makefile at {}: {e}", makefile.path);
    });
}
