use std::{fs, io::ErrorKind};
use regex::Regex;

const START_KEY: &str = "#=mgen_start=#\n";
const END_KEY: &str = "\n#=mgen_end=#";

#[derive(Debug)]
#[allow(dead_code)]
pub struct Makefile {
    pub path: String,
    pub prologue: String,
    pub mgen_zone: String,
    pub epilogue: String,
    pub include_dirs: Vec<String>,
}

impl Makefile {
    pub fn new_from_path(path: &str) -> Makefile {
        let contents = match fs::read_to_string(path) {
            Ok(v) => v,
            Err(e) => match e.kind() {
                // doesn't matter if the file doesn't exist, it will be created 
                // later.
                ErrorKind::NotFound => return Makefile { 
                    path: String::from(path), prologue: String::from(""),
                    mgen_zone: String::from(""), epilogue: String::from(""),
                    include_dirs: vec![String::from("")] },
                _ => panic!("error: cannot read makefile: {e}")
            }
        };

        let contains_start = contents.find(START_KEY).is_some();
        let contains_end = contents.find(END_KEY).is_some();

        if (contains_start && !contains_end) || (!contains_start && 
                                                                contains_end) {
            panic!("error: makefile contains one mgen guard but not the other. cannot continue.");
        }

        if !(contains_start && contains_end) {
            let include_dirs = Makefile::get_include_dirs(&contents);
            return Makefile { path: String::from(path), prologue: contents,
                mgen_zone: String::from(""), epilogue: String::from(""), include_dirs };
        }

        // contains the part before the start key and everything after
        let pre_and_after: Vec<&str> = contents.split(START_KEY).collect();
        // contains the space between start and end keys and also the space after
        let gen_and_post: Vec<&str> = pre_and_after[1].split(END_KEY).collect();

        let include_dirs = Makefile::get_include_dirs(&contents);
        
        Makefile { 
            path: String::from(path),
            prologue: String::from(pre_and_after[0]),
            mgen_zone: String::from(gen_and_post[0]),
            epilogue: String::from(gen_and_post[1]),
            include_dirs
        }
    }
    
    pub fn get_include_dirs(search: &String) -> Vec<String> {
        let mut include_dirs: Vec<String> = Vec::new();

        let makefile_regex = Regex::new(r"INCS\s*=\s*(.*)").unwrap();
        let caps = makefile_regex.captures_iter(search);
        let dirlist = match caps.map(|c| c.extract()).last() {
            Some((_, [v])) => v,
            None => {
                return include_dirs
            }
            ,
        };


        let incdir_regex = Regex::new(r"\s*-I\s*(.*?)(?:[\s]|$)").unwrap();
        let caps = incdir_regex.captures_iter(&dirlist);
        for (_, [dir]) in caps.map(|c| c.extract()) {
            include_dirs.push(String::from(dir));
        }


        include_dirs
    }
        
    pub fn contains_builddir_def(&self) -> bool {
        let re = Regex::new(r"\s+BUILDDIR = ").unwrap();
        re.is_match(&(self.prologue)) || re.is_match(&(self.epilogue))
    }

    pub fn contains_objdir_def(&self) -> bool {
        let re = Regex::new(r"\s+OBJDIR = ").unwrap();
        re.is_match(&(self.prologue)) || re.is_match(&(self.epilogue))
    }
     
    pub fn write_new_generated_content(&self, generated_content: String) -> std::io::Result<()> {
        let content = [&self.prologue, START_KEY, &generated_content, END_KEY, &self.epilogue].concat();

        fs::write(&self.path, &content)?;

        Ok(())
    }
}
