use std::fs;
use std::fs::DirEntry;

use crate::build_rule::*;
use crate::makefile::*;

#[derive(Debug)]
pub struct RuleList(pub Vec<BuildRuleKind<BuildRule>>);

impl RuleList {
    pub fn new() -> RuleList{
        RuleList(vec![])
    }

    pub fn recursively_gen_objects(&mut self, proj_dir_path: &str, makefile: &Makefile) -> std::io::Result<()>{
        let files = fs::read_dir(proj_dir_path)?;
        for file in files {
            let file = file?;
            let is_dir = file.metadata()?.is_dir();

            // ignore hidden files
            if file_is_hidden(&file) {
                continue;
            }

            let path = file.path();

            if is_dir {
                self.recursively_gen_objects(path.to_str().unwrap(), makefile)?;
            } else {
                match path.extension() {
                    Some(v) => {
                        let v = v.to_str().unwrap();
                        if v != "c" {
                            continue;
                        }
                        let path = path.to_str().unwrap();
                        self.0.push(BuildRuleKind::Object(BuildRule::new_from_c_file(path, makefile)));
                    },
                    None => {
                        continue;
                    },
                }
            }
        }
        Ok(())
    }

    pub fn get_obj_targets(&self) -> Vec<String>{
        let mut ret: Vec<String> = Vec::new();
        for rule in &self.0 {
            match rule {
                BuildRuleKind::Object(r) => ret.push(r.target.clone()),
                _ => break,
            }
        }
        ret
    }

    pub fn add_target_rule(&mut self) {
        let mut objects = self.get_obj_targets();
        let target_comm = String::from(
            format!("$(CC) $(LDFLAGS) $(INCS) -o $(BUILDDIR)/$(TARGET) {} $(LDLIBS)", objects.join(" "))
        );

        objects.push(String::from("| $(OBJDIR) $(BUILDDIR)"));
        let prereqs = objects;

        self.0.push(BuildRuleKind::Target(BuildRule {
            target: String::from("$(BUILDDIR)/$(TARGET)"),
            prerequisites: prereqs,
            recipe_commands: vec![target_comm],
        }))
    }

    pub fn add_directory_rule(&mut self, dir: &str) {
        self.0.push(BuildRuleKind::Directory(BuildRule {
            target: dir.to_string(),
            prerequisites: vec![],
            recipe_commands: vec![format!("mkdir -p {dir}").to_string()]
        }));
    }

    pub fn add_clean_rule(&mut self) {
        self.0.push(BuildRuleKind::Convenience(BuildRule {
            target: String::from(".PHONY: clean\nclean"),
            prerequisites: vec![],
            recipe_commands: vec![String::from("rm -r $(OBJDIR)"), String::from("rm -r $(BUILDDIR)")]
        }));
    }

    pub fn add_run_rule(&mut self) {
        self.0.push(BuildRuleKind::Convenience(BuildRule {
            target: String::from(".PHONY: run\nrun"),
            prerequisites: vec![String::from("$(BUILDDIR)/$(TARGET)")],
            recipe_commands: vec![String::from("./$(BUILDDIR)/$(TARGET)")]
        }));
    }

    pub fn get_target_rules(&self) -> Vec<BuildRuleKind<BuildRule>> {
        let mut ret: Vec<BuildRuleKind<BuildRule>> = Vec::new();
        for rule in &self.0 {
            match rule {
                BuildRuleKind::Target(r) => ret.push(BuildRuleKind::Target(r.clone())),
                _ => continue,
            }
        }
        ret
    }

    pub fn get_object_rules(&self) -> Vec<BuildRuleKind<BuildRule>> {
        let mut ret: Vec<BuildRuleKind<BuildRule>> = Vec::new();
        for rule in &self.0 {
            match rule {
                BuildRuleKind::Object(r) => ret.push(BuildRuleKind::Object(r.clone())),
                _ => continue,
            }
        }
        ret
    }

    pub fn get_directory_rules(&self) -> Vec<BuildRuleKind<BuildRule>> {
        let mut ret: Vec<BuildRuleKind<BuildRule>> = Vec::new();
        for rule in &self.0 {
            match rule {
                BuildRuleKind::Directory(r) => ret.push(BuildRuleKind::Directory(r.clone())),
                _ => continue,
            }
        }
        ret
    }

    pub fn get_convenience_rules(&self) -> Vec<BuildRuleKind<BuildRule>> {
        let mut ret: Vec<BuildRuleKind<BuildRule>> = Vec::new();
        for rule in &self.0 {
            match rule {
                BuildRuleKind::Convenience(r) => ret.push(BuildRuleKind::Convenience(r.clone())),
                _ => continue,
            }
        }
        ret
    }
}

fn file_is_hidden(file: &DirEntry) -> bool{
    let file_name = file.file_name();
    let str_name = file_name.to_str().unwrap();
    let first_char = str_name.chars().nth(0).unwrap();
    if first_char == '.' {
        true
    } else {
        false
    }
}

impl ToString for RuleList {
    fn to_string(&self) -> String {
        let targets = &self.get_target_rules().iter().map(|v| v.to_string()).collect::<Vec<_>>().join("\n");
        let objects = &self.get_object_rules().iter().map(|v| v.to_string()).collect::<Vec<_>>().join("\n");
        let directories = &self.get_directory_rules().iter().map(|v| v.to_string()).collect::<Vec<_>>().join("\n");
        let conveniences = &self.get_convenience_rules().iter().map(|v| v.to_string()).collect::<Vec<_>>().join("\n");

        let all = [targets, objects, directories, conveniences];
        let mut nonzero: Vec<String> = Vec::new();

        for category in all {
            if category.len() != 0 {
                nonzero.push(category.clone());
            }
        }
        
        nonzero.join("\n\n")
    }
}
